/* mpi_jet.c **********************************\
*                                              *
*     -- Turbulent jet modeled by LES --       *
*    explicit two-stage time differencing      *
*                                              *
*  written by Andrei Chernousov                *
*  E-mail: <andrei99@iname.com>                *
*  http://www.geocities.com/andrei_chernousov  *
*  Version 0.2, March 19, 2001                 *
\**********************************************/

#include "mpi.h"  // Message Passing Interface (MPI)

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

// continuation flag
int continFlag = 1;

//--- Combinations of the ratio of specific heats ---
#define K        1.4
#define K_1      0.4
#define _K_1   (-0.4)
#define _05K     0.7
#define K_K_1    3.5
#define K_1_2    0.2
#define _1_2_K_1 1.25

#define Cv      717.75 // heat capacity at constant volume, J/(kg*K)*/
#define R_VOZD  287.10 // gas constant
#define KR_VOZD 401.94 // gas constant * 1.4

#define Fx1 xU1 // aliases for x-fluxes
#define Fx2 xU2
#define Fx3 xU3
#define Fx4 xU4
#define Fx5 xU5

#define Fy1 yU1 // aliases for y-fluxes
#define Fy2 yU2
#define Fy3 yU3
#define Fy4 yU4
#define Fy5 yU5

#define Fz1 zU1 // aliases for z-fluxes
#define Fz2 zU2
#define Fz3 zU3
#define Fz4 zU4
#define Fz5 zU5

//=== Global variables
float
	// conservative variables at cells' centroids
	***U1,  ***U2,  ***U3,  ***U4,  ***U5,
	***U1p, ***U2p, ***U3p, ***U4p, ***U5p,
	// pointers for substitution
	***U1_, ***U2_, ***U3_, ***U4_, ***U5_,
	// conservative variables at cell boundaries in x-direction
	***xU1, ***xU2, ***xU3, ***xU4, ***xU5,
	***U1x, ***U2x, ***U3x, ***U4x, ***U5x,
	// conservative variables at cell boundaries in y-direction
	***yU1, ***yU2, ***yU3, ***yU4, ***yU5,
	***U1y, ***U2y, ***U3y, ***U4y, ***U5y,
	// conservative variables at cell boundaries in z-direction
	***zU1, ***zU2, ***zU3, ***zU4, ***zU5,
	***U1z, ***U2z, ***U3z, ***U4z, ***U5z,
	// primitive parameters
	R, U, V, W, P, C,
	// conservative variables
	u1, u2, u3, u4, u5,
	// cell spacings
	deltaX, deltaY, deltaZ,
	// time step
	deltaT,
	// ratios with deltas
	deltaT_X, deltaT_Y, deltaT_Z,
	_deltaX, _deltaY, _deltaZ,
	_2deltaX, _2deltaY, _2deltaZ,
	_4deltaX, _4deltaY, _4deltaZ,
	// array of probes
	*probes;

//--- MPI Buf
float *Buf;
unsigned BufCountU, BufCountF; // BufCountU > BufCountF

int
	step,    // current time step
	numstep, // overall prescribed number of time steps
	Answer,  // solution continuation flag
	f_step;  // frame taking step

char Stage;    // indicator of the current stage

unsigned
	LEN, HIG, DEP,  // cell numbers in x-, y-, z-directions, for 3D box
	LENN, HIGG, DEPP,

	mem,     // total amount of dynamic memory required for process (in bytes)

	i,  j,  k,
	i__, j__, k__, // indices
	l;

//--- handles of "probes.myid" files
FILE *pFprobes;

//--- for the specification of disturbed velocity
float
	Vd, Wd, // curent disturbed velocities
	Va, Wa; // their amplitude values
int
	Ns_min, // minimal period of existence of this disturbing block
	Nst,    //   and its max-min
	counter = 0;

//--- rough circle
int j_c, k_c;
float ly, lz, L, Radius;

//--- buffer string
char str[80];

//--- for LCS approximate Riemann solver ---
float
	 P_, _P, R_, _R, U_, _U, V_, _V, W_, _W, C_, _C, _T, T_,
	 C_p, C_m, C_o, C_p_, C_m_, C_o_,
	 _jo, _jp, _jm, jo_, jp_, jm_,
	 al_p, al_m, al_o,
	 J_p, J_m, J_o,

//--- for gradient fluxes ---
	 // matrix of derivatives of velocity
	 du_dx, dv_dx, dw_dx,
	 du_dy, dv_dy, dw_dy,
	 du_dz, dv_dz, dw_dz,
	 iu, ui, ju, uj, ku, uk,
	 iv, vi, jv, vj, kv, vk,
	 iw, wi, jw, wj, kw, wk,
	 iU, Ui, jU, Uj, kU, Uk,
	 iV, Vi, jV, Vj, kV, Vk,
	 iW, Wi, jW, Wj, kW, Wk,
	 // mean filtered velocity components and density at the interfaces
	 uu, vv, ww, rr,
	 // viscous stress tensor
	 sigma_xx, sigma_xy, sigma_xz,
	 sigma_yx, sigma_yy, sigma_yz,
	 sigma_zx, sigma_zy, sigma_zz,
	 // heat flux vector
	 q_x, q_y, q_z,
	 //-- transport coefficients
	 // molecular
	 mu_L, lambda_L, Pr_L,
	 // subgrid-scale
	 mu_T, Pr_T, cp_Pr_T,
	 // effective
	 mu_E, lambda_E,
	 // Smagorinsky constant, complex CsDD = Cs * delta * delta, where delta = (dX*dY*dZ)~0.33333
	 Cs, CsDD,
	 // invariant of deforming rate tensor
	 _S_;

//--- for monotone piecevise parabolic reconstruction ---
float
	// finite differences of conservative variables
	m1, m2, m3, m4, m5,
	w1, w2, w3, w4, w5,
	// finite differences of characteristic variables
	M1, M2, M3, M4, M5,
	W1, W2, W3, W4, W5,
	// modified finite differences of characteristic variables
	_M1, _M2, _M3, _M4, _M5,
	_W1, _W2, _W3, _W4, _W5,
	// transformation matrix [S-1]
	S11, S12, S13,
				 S23,
	S41, S42,
	S51, S52,
	// inverted transformation matrix [S-1]
	_S23, _S24, _S25,
	_S33,
	_S43,
	_S53, _S54, _S55,
	// subsidiary
	aa, bb, cc, dd, ee, ff, gg, hh, ll, mm, nn,
	u_hh, v_hh, w_hh, u_ee, v_ee, w_ee,
	k1, k2, k3, k4, k5;

	// constant coefficients
#define BB 4.000000000000000
#define C1 0.333333333333333
#define C2 0.166666666666667

float ay;

/***********
*  MINMOD  *   Limiter function minmod(x,y) - Chakravarthy, Osher
***********/ /*inline*/
float minmod(float x, float y) {

	if(x*y <= 0.)
		return 0.;

	ay = ( (y*=BB) > 0. ) ? y: -y;

	if (x > 0.)
		return (x<ay) ? x : ay;

	return ( (-x<ay) ? x: -ay );

} // end minmod()

//--- for dynamic allocation of memory ---

/************
*  ARRAY2D  *   Memory allocation procedure for 2D arrays
************/
float **Array2D(unsigned columns, unsigned rows)
{
float **x;
unsigned i;

	if((x=(float **)malloc(columns*sizeof(float*))) == NULL) {
		fprintf(stderr, "mpi_duct: can't allocate memory");
		MPI_Abort(MPI_COMM_WORLD, 1);
	}
	for (i = 0; i < columns; i++)
		if ((x[i]=(float *)malloc(rows*sizeof(float))) == NULL) {
			fprintf(stderr, "mpi_duct: can't allocate memory");
			MPI_Abort(MPI_COMM_WORLD, 1);
		}
	return x;

} // end Array2D()

/************
*  ARRAY3D  *   Memory allocation procedure for 3D arrays
************/
float ***Array3D(unsigned columns, unsigned rows, unsigned floors)
{
float ***x;
unsigned i, j;

	if( (x=(float ***)malloc(columns*sizeof(float**))) == NULL ) {
		fprintf(stderr, "mpi_duct: can't allocate memory");
		MPI_Abort(MPI_COMM_WORLD, 1);
	}
	for (i = 0; i < columns; i++) {
		if ((x[i]=(float **)malloc(rows*sizeof(float*))) == NULL) {
			fprintf(stderr, "mpi_duct: can't allocate memory");
			MPI_Abort(MPI_COMM_WORLD, 1);
		}
		for (j = 0; j < rows; j++)
			if ((x[i][j]=(float *)malloc(floors*sizeof(float))) == NULL) {
				fprintf(stderr, "mpi_duct: can't allocate memory");
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
	}
	return x;

} // end Array3D()


/***************
*  INITIALIZE  *    all necessary initializstions
***************/
void Initialize(int myid) // identifier of _this_ process
{
  // temporary arrays
float ***fp[10];
char filename[30];

unsigned i, jj, kk;
MPI_File fh;
MPI_Status status;
float buf;


	//--- root process (node w/rank 0) reads data from file "mpi_duct.bin"
	//    and broadcasts it to others
	if(0 == myid) {
		// open the file...
		MPI_File_open(MPI_COMM_SELF, "mpi_jet.bin", MPI_MODE_CREATE | MPI_MODE_RDWR, MPI_INFO_NULL, &fh);
		MPI_File_set_view(fh, 0, MPI_FLOAT, MPI_FLOAT, "native", MPI_INFO_NULL);
	}

	//--- ... and reads/broadcast data (if myid == 0), otherwise receives it
	i = 0;
	do{

		// syncronization
		MPI_Barrier(MPI_COMM_WORLD);

		// root reads the next data item
		if (0 == myid) MPI_File_read(fh, &buf, 1, MPI_FLOAT, &status);

		// root process broadcasts the next data item
		MPI_Bcast(&buf, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
		fprintf(stdout, "%d process: after broadcast %dth data item = %f\n",
					myid+1, i, buf);

		// dispatch to specific parameters
		switch (i) {
			case  0: fprintf(stdout, "LEN    = %d\n", LEN    = buf); break;
			case  1: fprintf(stdout, "HIG    = %d\n", HIG    = buf); break;
			case  2: fprintf(stdout, "DEP    = %d\n", DEP    = buf); break;
			case  3: fprintf(stdout, "deltaX = %f\n", deltaX = buf); break;
			case  4: fprintf(stdout, "deltaY = %f\n", deltaY = buf); break;
			case  5: fprintf(stdout, "deltaZ = %f\n", deltaZ = buf); break;
			case  6: fprintf(stdout, "deltaT = %f\n", deltaT = buf); break;
			case  7: fprintf(stdout, "Answer = %d\n", Answer = buf); break;
			case  8: fprintf(stdout, "numstep= %d\n", numstep= buf); break;
			case  9: fprintf(stdout, "mu_L   = %f\n", mu_L   = buf); break;
			case 10: fprintf(stdout, "Pr_L   = %f\n", Pr_L   = buf); break;
			case 11: fprintf(stdout, "Pr_T   = %f\n", Pr_T   = buf); break;
			case 12: fprintf(stdout, "Cs     = %f\n", Cs     = buf); break;
			case 13: fprintf(stdout, "Va     = %f\n", Va     = buf); break;
			case 14: fprintf(stdout, "Wa     = %f\n", Wa     = buf); break;
			case 15: fprintf(stdout, "Ns_min = %f\n", Ns_min = buf); break;
			case 16: fprintf(stdout, "Nst    = %d\n", Nst    = buf); break;
			case 17: fprintf(stdout, "f_step = %d\n", f_step = buf); break;
			case 18: fprintf(stdout, "j_c    = %d\n", j_c    = buf); break;
			case 19: fprintf(stdout, "k_c    = %d\n", k_c    = buf); break;
			case 20: fprintf(stdout, "Radius = %f\n", Radius = buf); break;
			default: break;
		} // end switch

		i++;

	} while (i < 21);

	//--- close file
	if (0 == myid) MPI_File_close(&fh);

	//--- other initializatons
	// complexes with deltas
	deltaT_X = deltaT / deltaX;
	deltaT_Y = deltaT / deltaY;
	deltaT_Z = deltaT / deltaZ;
	 _deltaX = 1. / deltaX;
	 _deltaY = 1. / deltaY;
	 _deltaZ = 1. / deltaZ;
	_4deltaX = 1. / ( 4 * deltaX );
	_4deltaY = 1. / ( 4 * deltaY );
	_4deltaZ = 1. / ( 4 * deltaZ );
	_2deltaX = _4deltaX + _4deltaX;
	_2deltaY = _4deltaY + _4deltaY;
	_2deltaZ = _4deltaZ + _4deltaZ;
	// box sizes + 1
	LENN = LEN + 1;
	HIGG = HIG + 1;
	DEPP = DEP + 1;
	// constant-value lambda_L
	lambda_L = mu_L * Cv * K / Pr_L;
	// complex cp_Pr_T
	cp_Pr_T = Cv * K / Pr_T;
	// complex Cd * delta * delta in Smagorinsky model
	CsDD = Cs * pow(deltaX * deltaY * deltaZ, 0.6666666666667);

	//--- dynamic allocation of memory
	// total amount of memory required, for gas dynamics 5
	mem = 10 * (LEN + 2) * (HIG + 2) * (DEP + 2)
		 + 10 * (LEN + 1) *  HIG	   *  DEP
		 + 10 *  LEN      * (HIG + 1) *  DEP
		 + 10 *  LEN      *  HIG	   * (DEP + 1);
	fprintf(stdout, "%d process: %d bytes of memory required\n",
				myid+1, mem*sizeof(float) );
	// for quantities in cells
	for (i = 0; i < 10; i++) fp[i] = Array3D(LEN+2, HIG+2, DEP+2);
	U1 = fp[0];	U1p= fp[5];
	U2 = fp[1]; U2p= fp[6];
	U3 = fp[2]; U3p= fp[7];
	U4 = fp[3]; U4p= fp[8];
	U5 = fp[4]; U5p= fp[9];
	// for x fluxes
	for (i = 0; i < 10; i++) fp[i] = Array3D(LENN, HIG, DEP);
	xU1 = fp[0]; U1x = fp[5];
	xU2 = fp[1]; U2x = fp[6];
	xU3 = fp[2]; U3x = fp[7];
	xU4 = fp[3]; U4x = fp[8];
	xU5 = fp[4]; U5x = fp[9];
	// for y fluxes
	for (i = 0; i < 10; i++) fp[i] = Array3D(LEN, HIGG, DEP);
	yU1 = fp[0]; U1y = fp[5];
	yU2 = fp[1]; U2y = fp[6];
	yU3 = fp[2]; U3y = fp[7];
	yU4 = fp[3]; U4y = fp[8];
	yU5 = fp[4]; U5y = fp[9];
	// for z fluxes
	for (i = 0; i < 10; i++) fp[i] = Array3D( LEN, HIG, DEPP );
	zU1 = fp[0]; U1z = fp[5];
	zU2 = fp[1]; U2z = fp[6];
	zU3 = fp[2]; U3z = fp[7];
	zU4 = fp[3]; U4z = fp[8];
	zU5 = fp[4]; U5z = fp[9];
	//
	fprintf(stdout, "%d process: 3D arrays allocated\n", myid+1);

		// MPI Buf
	BufCountF = 5 *  HIG    *  DEP;    // BufCountF < BufCountU
	BufCountU = 5 * (HIG+2) * (DEP+2); // BufCountU
	if((Buf = (float *)malloc(BufCountU * sizeof(float))) == NULL) {
		fprintf(stderr, "mpi_jet: can't allocate memory.\n");
		MPI_Abort(MPI_COMM_WORLD, 1);
	}

		/* array of probes */
	if((probes=(float *)malloc(sizeof(float) * HIG)) == NULL) {
		fprintf(stderr, "mpi_jet: can't allocate memory for probes.\n");
		MPI_Abort(MPI_COMM_WORLD, 1);
	}

	//--- Initial conditions
	// start new simulation
	if(0 == Answer) {
		//
		P = 100000.;
		R = 1.0;
		U = 100.;
		//
		for (i = 1; i < LENN; i++) {
			for (j = 1; j <= HIG; j++) {
				for (k = 1; k < DEPP; k++) {
					U1[i][j][k] = R;
					U2[i][j][k] = R * U;
					U3[i][j][k] = 0.;
					U4[i][j][k] = 0.;
					U5[i][j][k] = P / K_1 + 0.5 * R * U * U;
				}
			}
		}
		// from the beginning
		step = 0;
	}

	// continue previously saved simulation
	else {
		//
		sprintf(filename, "backup.%d", myid);
		MPI_File_open(MPI_COMM_SELF, filename, MPI_MODE_CREATE | MPI_MODE_RDWR,
							MPI_INFO_NULL, &fh);
		MPI_File_set_view(fh, 0, MPI_FLOAT, MPI_FLOAT, "native", MPI_INFO_NULL);
		//
		for (i = 1; i < LENN; i++) {
			for (j = 1; j < HIGG; j++) {
				MPI_File_read(fh, &U1[i][j][1], DEP, MPI_FLOAT, &status);
				MPI_File_read(fh, &U2[i][j][1], DEP, MPI_FLOAT, &status);
				MPI_File_read(fh, &U3[i][j][1], DEP, MPI_FLOAT, &status);
				MPI_File_read(fh, &U4[i][j][1], DEP, MPI_FLOAT, &status);
				MPI_File_read(fh, &U5[i][j][1], DEP, MPI_FLOAT, &status);
			}
		}
	    	// restore current step		
		MPI_File_read(fh, &buf, 1, MPI_FLOAT, &status);
		step = (int)buf;
	    	// restore and set the seed of pseudo-random number
		MPI_File_read(fh, &buf, 1, MPI_FLOAT, &status);		
	    	srand((unsigned int)buf);
		// close bakup file
		MPI_File_close(&fh);
	}

	// create/open "probes.muid" file in binary mode
	sprintf(filename, "probes.%d", myid);
	if ((pFprobes = fopen(filename, "ab")) == NULL)
		fprintf(stdout, "can't open \"%s\".\n", filename);

	// first stage
	Stage = 1;
	U1_ = U1;
	U2_ = U2;
	U3_ = U3;
	U4_ = U4;
	U5_ = U5;
	fprintf(stdout, "%d process: after initialization.\n", myid+1);

} // end Initialize()


/***********************
* BOUNCONDINGHOSTCELLS *   BC in the cells out of the domain
***********************/
void BounCondInGhostCells(
	int myid,      // identifier of _this_ process
	int numprocs ) // number of processes
{
unsigned i, j, k;

int next, previous; // processes next/previous to this
MPI_Status status;

	//--- Communicate via MPI to complete cell x-layers ---

	//-- sending to next processes

	// next process's ID
	next = myid+1;

	// root process
	if (0 == myid) {
		// sending to the next
		for (i = 0, j = 0; j <= HIGG; j++) {
			for (k = 0; k <= DEPP; k++, i += 5) {
				Buf[i  ] = U1_[LEN][j][k];
				Buf[i+1] = U2_[LEN][j][k];
				Buf[i+2] = U3_[LEN][j][k];
				Buf[i+3] = U4_[LEN][j][k];
				Buf[i+4] = U5_[LEN][j][k];
			}
		}
		MPI_Send(Buf, BufCountU, MPI_FLOAT, next, 99, MPI_COMM_WORLD);
		// previous doesn't exist
/*		MPI_Recv(Buf, BufCountU, MPI_FLOAT, MPI_ANY_SOURCE, 99, MPI_COMM_WORLD,
					&status);
		for (i = 0, j = 0; j <= HIGG; j++) {
			for (k = 0; k <= DEPP; k++, i += 5) {
				U1_[0][j][k] = Buf[i  ];
				U2_[0][j][k] = Buf[i+1];
				U3_[0][j][k] = Buf[i+2];
				U4_[0][j][k] = Buf[i+3];
				U5_[0][j][k] = Buf[i+4];
			}
		}
*/
	} // end if()

	// other processes
	else {
		// receiving from the previous, everyone
		MPI_Recv(Buf, BufCountU, MPI_FLOAT, MPI_ANY_SOURCE, 99, MPI_COMM_WORLD,
					&status);
		for (i = 0, j = 0; j <= HIGG; j++) {
			for (k = 0; k <= DEPP; k++, i += 5) {
				U1_[0][j][k] = Buf[i  ];
				U2_[0][j][k] = Buf[i+1];
				U3_[0][j][k] = Buf[i+2];
				U4_[0][j][k] = Buf[i+3];
				U5_[0][j][k] = Buf[i+4];
			}
		}
		// sending to the next; if not the last!
		if(myid != numprocs-1) {
			for (i = 0, j = 0; j <= HIGG; j++) {
				for (k = 0; k <= DEPP; k++, i += 5) {
					Buf[i  ] = U1_[LEN][j][k];
					Buf[i+1] = U2_[LEN][j][k];
					Buf[i+2] = U3_[LEN][j][k];
					Buf[i+3] = U4_[LEN][j][k];
					Buf[i+4] = U5_[LEN][j][k];
				}
			}
			MPI_Send(Buf, BufCountU, MPI_FLOAT, next, 99, MPI_COMM_WORLD);
		} // end if(myid != numprocs-1)
	}

	//-- sending to previous processes

	// previous process's ID
	previous = myid-1;

	// root process
	if(0 == myid) {
		// previous doesn't exist
/*		for (i = 0, j = 0; j <= HIGG; j++) {
			for (k = 0; k <= DEPP; k++, i += 5) {
				Buf[i  ] = U1_[1][j][k];
				Buf[i+1] = U2_[1][j][k];
				Buf[i+2] = U3_[1][j][k];
				Buf[i+3] = U4_[1][j][k];
				Buf[i+4] = U5_[1][j][k];
			}
		}
		MPI_Send(Buf, BufCountU, MPI_FLOAT, previous, 99, MPI_COMM_WORLD);
*/
		// receiving from the next
		MPI_Recv(Buf, BufCountU, MPI_FLOAT, MPI_ANY_SOURCE, 99, MPI_COMM_WORLD,
					&status);
		for (i = 0, j = 0; j <= HIGG; j++) {
			for (k = 0; k <= DEPP; k++, i += 5) {
				U1_[LENN][j][k] = Buf[i  ];
				U2_[LENN][j][k] = Buf[i+1];
				U3_[LENN][j][k] = Buf[i+2];
				U4_[LENN][j][k] = Buf[i+3];
				U5_[LENN][j][k] = Buf[i+4];
			}
		}
	} // end if()

	// other processes
	else {
		// receiving from the next; if not the last process!
		if(myid != numprocs-1) {
			MPI_Recv(Buf, BufCountU, MPI_FLOAT, MPI_ANY_SOURCE, 99,
						MPI_COMM_WORLD, &status);
			for (i = 0, j = 0; j <= HIGG; j++) {
				for (k = 0; k <= DEPP; k++, i += 5) {
					U1_[LENN][j][k] = Buf[i  ];
					U2_[LENN][j][k] = Buf[i+1];
					U3_[LENN][j][k] = Buf[i+2];
					U4_[LENN][j][k] = Buf[i+3];
					U5_[LENN][j][k] = Buf[i+4];
				}
			}
		} // end if(myid != numprocs-1)
		// sending to previous - everyone
		for (i = 0, j = 0; j <= HIGG; j++) {
			for (k = 0; k <= DEPP; k++, i += 5) {
				Buf[i  ] = U1_[1][j][k];
				Buf[i+1] = U2_[1][j][k];
				Buf[i+2] = U3_[1][j][k];
				Buf[i+3] = U4_[1][j][k];
				Buf[i+4] = U5_[1][j][k];
			}
		}
		MPI_Send(Buf, BufCountU, MPI_FLOAT, previous, 99, MPI_COMM_WORLD);
	}

	// syncronization
	MPI_Barrier(MPI_COMM_WORLD);


	//-- x
	// left side - JET out of a cirular area
	if(0 == myid) { /* root process */
		P = 100000.;
		R = 1.0;
		for (j = 1; j < HIGG; j++) {
			for (k = 1; k < DEPP; k++) {
				ly = deltaY*((float)j-(float)j_c);
				lz = deltaZ*((float)k-(float)k_c);
				L  = sqrt(ly*ly + lz*lz);
				if (L <= Radius){
					U = 200.;
					V = Vd;
					W = Wd;
				}
				else {
					U = 50.;
					V=W=0.;
				}
				U1_[0][j][k] = R;
				U2_[0][j][k] = R * U;
				U3_[0][j][k] = R * V;
				U4_[0][j][k] = R * W;
				U5_[0][j][k] = P / K_1 + 0.5 * R * (U * U + V * V + W * W);
			}
		}
	} // end if(0 == myid)

	// right side - OUTFLOW
	if(myid == numprocs-1) { /* last process */
		P = 100000.;
		for( j = 1; j < HIGG; j++ ) {
			for( k = 1; k < DEPP; k++ ) {
				U1_[LENN][j][k] = R = U1_[LEN][j][k];
				U2_[LENN][j][k] = U = U2_[LEN][j][k];
				U3_[LENN][j][k] = V = U3_[LEN][j][k];
				U4_[LENN][j][k] = W = U4_[LEN][j][k];
				U5_[LENN][j][k] = P / K_1 + 0.5 * ( U * U + V * V + W * W ) / R;
			}
		}
	} // end if(myid == numprocs-1)

	//-- z - SLIP
	for( i = 0; i <= LENN; i++ ) {
		for( j = 1; j < HIGG; j++ ) {
			// back side
			U1_[i][j][0] =   U1_[i][j][DEP];
			U2_[i][j][0] =   U2_[i][j][DEP];
			U3_[i][j][0] =   U3_[i][j][DEP];
			U4_[i][j][0] = - U4_[i][j][DEP];
			U5_[i][j][0] =   U5_[i][j][DEP];
			// front side
			U1_[i][j][DEPP] =   U1_[i][j][1];
			U2_[i][j][DEPP] =   U2_[i][j][1];
			U3_[i][j][DEPP] =   U3_[i][j][1];
			U4_[i][j][DEPP] = - U4_[i][j][1];
			U5_[i][j][DEPP] =   U5_[i][j][1];
		}
	}

	//-- y - SLIP
	for( i = 0; i <= LENN; i++ ) {
		for( k = 0; k <= DEPP; k++ ) {
			// bottom side
			U1_[i][0][k] =   U1_[i][1][k];
			U2_[i][0][k] =   U2_[i][1][k];
			U3_[i][0][k] = - U3_[i][1][k];
			U4_[i][0][k] =   U4_[i][1][k];
			U5_[i][0][k] =   U5_[i][1][k];
			// top side
			U1_[i][HIGG][k] =   U1_[i][HIG][k];
			U2_[i][HIGG][k] =   U2_[i][HIG][k];
			U3_[i][HIGG][k] = - U3_[i][HIG][k];
			U4_[i][HIGG][k] =   U4_[i][HIG][k];
			U5_[i][HIGG][k] =   U5_[i][HIG][k];
		}
	}

} // end BounCondInGhostCells()


/*******************   piecewise-parabolic reconstruction parameters
*  RECONSTRUCTION  *   in cells with 2..3 order in space
*******************/
void Reconstruction(void)
{
float kk;
unsigned i, j, k, _i, _j, _k, i_, j_, k_;

	//
	for( i = 1, _i = 0, i_ = 2; i < LENN; i++, _i++, i_++ ) {
		for( j = 1, _j = 0, j_ = 2; j < HIGG; j++, _j++, j_++ ) {
			for( k = 1, _k = 0, k_ = 2; k < DEPP; k++, _k++, k_++ ) {
				/*---  X  ---*/
					/* at cell centroid */
				u1 = U1_[i][j][k];
				u2 = U2_[i][j][k];
				u3 = U3_[i][j][k];
				u4 = U4_[i][j][k];
				u5 = U5_[i][j][k];
				R =   1. / u1;   /* 1 / R  */
				U =   u2 * R;
				V =   u3 * R;
				W =   u4 * R;
				P = ( u5 - 0.5 * ( u2 * u2 + u3 * u3 + u4 * u4 ) * R ) * K_1;
				C = sqrt( K * P * R );
				/* evaluation of matrices */
				ff = U*U+V*V+W*W;   /*  */
				aa = K_1_2*ff;      /* (K-1)*(U*U+V*V+W*W)/2 */
				bb = _K_1*U;        /* - (K_1) * U           */
				cc = _K_1*V;        /* - (K_1) * V           */
				dd = _K_1*W;        /* - (K_1) * W           */
				ll = 1./(C+C);      /* 1/(2.*C)              */
				ee = ll/C;          /* 1./(2*C*C)            */
				gg = - ff * ee;     /* -(U*U+V*V+W*W)/(2*C*C)*/
				hh = - ( ee + ee ); /* - 1 / (C*C)           */
				mm = U*ll;          /* U/(2.*C)              */
				/* differencing the conservative variables */
					/* forvard */
				m1 = U1_[i_][j][k] - u1;
				m2 = U2_[i_][j][k] - u2;
				m3 = U3_[i_][j][k] - u3;
				m4 = U4_[i_][j][k] - u4;
				m5 = U5_[i_][j][k] - u5;
					/* backward */
				w1 = u1 - U1_[_i][j][k];
				w2 = u2 - U2_[_i][j][k];
				w3 = u3 - U3_[_i][j][k];
				w4 = u4 - U4_[_i][j][k];
				w5 = u5 - U5_[_i][j][k];
				/* transformation matrix [S] */
				S11 = aa - C*C;
				S41 = aa - (kk=C*U);   /* CU */
				S51 = aa + kk;
				S12 = S42 = S52 = bb;
				S42 += C;
				S52 -= C;
				/* finite differences of characteristic variables dW = S * dU */
					/* forward */
				kk = K_1 * m5;
				k2 = cc * m3;
				k3 = dd * m4;
				M1 = S11*m1 + S12*m2 +   k2 +   k3 + kk;
				M2 =  -V*m1 		 +   m3            ;
				M3 =  -W*m1 		 		+ 	m4     ;
				M4 = S41*m1 + S42*m2 +   k2 +   k3 + kk;
				M5 = S51*m1 + S52*m2 +   k2 +   k3 + kk;
					/* backward */
				kk = K_1 * w5;
				k2 = cc * w3;
				k3 = dd * w4;
				W1 = S11*w1 + S12*w2 +   k2 +   k3 + kk;
				W2 =  -V*w1 		 +   w3 	       ;
				W3 =  -W*w1 				+   w4     ;
				W4 = S41*w1 + S42*w2 +   k2 +   k3 + kk;
				W5 = S51*w1 + S52*w2 +   k2 +   k3 + kk;
				/* modified characteristic differences _W */
				_M1 = minmod( M1, W1 ); /* the best way */
				_M2 = minmod( M2, W2 );
				_M3 = minmod( M3, W3 );
				_M4 = minmod( M4, W4 );
				_M5 = minmod( M5, W5 );
				_W1 = minmod( W1, M1 );
				_W2 = minmod( W2, M2 );
				_W3 = minmod( W3, M3 );
				_W4 = minmod( W4, M4 );
				_W5 = minmod( W5, M5 );
				/* inverted transformation matrix [S-1] */
				/*_S21 = */ u_hh = U*hh; /*_S31 =*/ v_hh = V*hh; /*_S41 =*/ w_hh = W*hh;
								u_ee = U*ee; /*_S34 =*/ v_ee = V*ee; /*_S44 =*/ w_ee = W*ee;
				 _S24 = u_ee + ll; _S25 = u_ee - ll;
				_S54 = _S55 = ( nn = - 0.5 * gg ) + _1_2_K_1;
				_S54 += mm; _S55 -= mm;
				/* parameters' vector at the rightmost interface */
				k1 = _M1 * C1 + _W1 * C2;
				k2 = _M2 * C1 + _W2 * C2;
				k3 = _M3 * C1 + _W3 * C2;
				k4 = _M4 * C1 + _W4 * C2;
				k5 = _M5 * C1 + _W5 * C2;
				kk = k4 + k5;
				xU1[ i][_j][_k] = u1 +   hh * k1 + 						 	ee  * kk;
				xU2[ i][_j][_k] = u2 + u_hh * k1 + 						   _S24 * k4 + _S25 * k5;
				xU3[ i][_j][_k] = u3 + v_hh * k1 + 		  k2 			 + v_ee * kk;
				xU4[ i][_j][_k] = u4 + w_hh * k1 + 					  k3 + w_ee * kk;
				xU5[ i][_j][_k] = u5 +   gg * k1 +    V * k2 +    W * k3 + _S54 * k4 + _S55 * k5;
				/* ... leftmost interface */
				k1 = _M1 * C2 + _W1 * C1;
				k2 = _M2 * C2 + _W2 * C1;
				k3 = _M3 * C2 + _W3 * C1;
				k4 = _M4 * C2 + _W4 * C1;
				k5 = _M5 * C2 + _W5 * C1;
				kk =  k4 + k5;
				U1x[_i][_j][_k] = u1 -   hh * k1 - 						     ee * kk;
				U2x[_i][_j][_k] = u2 - u_hh * k1 - 						   _S24 * k4 - _S25 * k5;
				U3x[_i][_j][_k] = u3 - v_hh * k1 - 		  k2 			 - v_ee * kk;
				U4x[_i][_j][_k] = u4 - w_hh * k1 - 					  k3 - w_ee * kk;
				U5x[_i][_j][_k] = u5 -   gg * k1 -    V * k2 -    W * k3 - _S54 * k4 - _S55 * k5;
				/*---  Y  ---*/
				/* evaluation of matrices */
				kk = bb; bb = cc; cc = dd; dd = kk;
				mm = V*ll;          /* V/(2.*C) */
				/* differencing of the conservative variables */
					/* forvard */
				m1 = U1_[i][j_][k] - u1;
				m2 = U3_[i][j_][k] - u3; /* U3[i][j+1][k] ( U->V, V->W, W->U : 3-4-2 ) */
				m3 = U4_[i][j_][k] - u4; /* U4[i][j+1][k] */
				m4 = U2_[i][j_][k] - u2; /* U2[i][j+1][k] */
				m5 = U5_[i][j_][k] - u5;
					/* backward */
				w1 = u1 - U1_[i][_j][k];
				w2 = u3 - U3_[i][_j][k]; /* U3[i][j-1][k] ( U->V, V->W, W->U : 3-4-2 ) */
				w3 = u4 - U4_[i][_j][k]; /* U4[i][j-1][k] */
				w4 = u2 - U2_[i][_j][k]; /* U2[i][j-1][k] */
				w5 = u5 - U5_[i][_j][k];
				/* transformation matrix [S] */
				S41 = aa - (kk=C*V);    /* CV */
				S51 = aa + kk;
				S12 = S42 = S52 = bb;
				S42 += C;
				S52 -= C;
				/* finite differences of characteristic variables dW = S * dU */
					/* forward */
				kk = K_1 * m5;
				k2 = cc * m3;
				k3 = dd * m4;
				M1 = S11*m1 + S12*m2 +   k2 +   k3 + kk;
				M2 =  -W*m1 		 +   m3            ;  /* -W*m1 ! */
				M3 =  -U*m1 		 		+ 	m4     ;  /* -U*m1 ! */
				M4 = S41*m1 + S42*m2 +   k2 +   k3 + kk;
				M5 = S51*m1 + S52*m2 +   k2 +   k3 + kk;
					/* backward */
				kk = K_1 * w5;
				k2 = cc * w3;
				k3 = dd * w4;
				W1 = S11*w1 + S12*w2 +   k2 +   k3 + kk;
				W2 =  -W*w1 		 +   w3 	       ;  /* -W*m1 ! */
				W3 =  -U*w1 				+   w4     ;  /* -U*m1 ! */
				W4 = S41*w1 + S42*w2 +   k2 +   k3 + kk;
				W5 = S51*w1 + S52*w2 +   k2 +   k3 + kk;
				/* modified characteristic differences _W */
				_M1 = minmod( M1, W1 );
				_M2 = minmod( M2, W2 );
				_M3 = minmod( M3, W3 );
				_M4 = minmod( M4, W4 );
				_M5 = minmod( M5, W5 );
				_W1 = minmod( W1, M1 );
				_W2 = minmod( W2, M2 );
				_W3 = minmod( W3, M3 );
				_W4 = minmod( W4, M4 );
				_W5 = minmod( W5, M5 );
				/* inverted transformation matrix [S-1] */
				/*_S21 =  v_hh; 	  _S31 =  w_hh; _S41 =  u_hh; */
									/*_S34 =  w_ee; _S44 =  u_ee; */
				_S24 = v_ee + ll; _S25 = v_ee - ll;
				_S54 = _S55 = nn + _1_2_K_1;
				_S54 += mm; _S55 -= mm;
				/* parameters' vector at the upper interface */
				k1 = _M1 * C1 + _W1 * C2;
				k2 = _M2 * C1 + _W2 * C2;
				k3 = _M3 * C1 + _W3 * C2;
				k4 = _M4 * C1 + _W4 * C2;
				k5 = _M5 * C1 + _W5 * C2;
				kk = k4 + k5;
				yU1[_i][ j][_k] = u1 +   hh * k1 + 						 	ee  * kk;
				yU3[_i][ j][_k] = u3 + v_hh * k1 + 						   _S24 * k4 + _S25 * k5;
				yU4[_i][ j][_k] = u4 + w_hh * k1 + 		  k2 			 + w_ee * kk;
				yU2[_i][ j][_k] = u2 + u_hh * k1 + 					  k3 + u_ee * kk;
				yU5[_i][ j][_k] = u5 +   gg * k1 +    W * k2 +    U * k3 + _S54 * k4 + _S55 * k5;
				/* ...lower interface */
				k1 = _M1 * C2 + _W1 * C1;
				k2 = _M2 * C2 + _W2 * C1;
				k3 = _M3 * C2 + _W3 * C1;
				k4 = _M4 * C2 + _W4 * C1;
				k5 = _M5 * C2 + _W5 * C1;
				kk =  k4 + k5;
				U1y[_i][_j][_k] = u1 -   hh * k1 - 						     ee * kk;
				U3y[_i][_j][_k] = u3 - v_hh * k1 - 						   _S24 * k4 - _S25 * k5;
				U4y[_i][_j][_k] = u4 - w_hh * k1 - 		  k2 			 - w_ee * kk;
				U2y[_i][_j][_k] = u2 - u_hh * k1 - 					  k3 - u_ee * kk;
				U5y[_i][_j][_k] = u5 -   gg * k1 -    W * k2 -    U * k3 - _S54 * k4 - _S55 * k5;
				/*--- Z --- */
				/* evaluation of matrices */
				kk = bb; bb = cc; cc = dd; dd = kk;
				mm = W*ll;          /* W/(2.*C) */
				/*  differencing of the conservative variables */
				/* forward  */
				m1 = U1_[i][j][k_] - u1;
				m2 = U4_[i][j][k_] - u4; /* U4[i][j][k+1] ( V->W, W->U, U->V : 4-2-3 ) */
				m3 = U2_[i][j][k_] - u2; /* U2[i][j][k+1] */
				m4 = U3_[i][j][k_] - u3; /* U3[i][j][k+1] */
				m5 = U5_[i][j][k_] - u5;
				/* backward */
				w1 = u1 - U1_[i][j][_k];
				w2 = u4 - U4_[i][j][_k]; /* U4[i][j][k-1] ( V->W, W->U, U->V : 4-2-3 ) */
				w3 = u2 - U2_[i][j][_k]; /* U2[i][j][k-1] */
				w4 = u3 - U3_[i][j][_k]; /* U3[i][j][k-1] */
				w5 = u5 - U5_[i][j][_k];
				/* transformation matrix [S] */
				S41 = aa - (kk=C*W);    /* CW */
				S51 = aa + kk;
				S12 = S42 = S52 = bb;
				S42 += C;
				S52 -= C;
				/* finite differences of characteristic variables dW = S * dU */
					/* forward */
				kk = K_1 * m5;
				k2 = cc * m3;
				k3 = dd * m4;
				M1 = S11*m1 + S12*m2 +   k2 +   k3 + kk;
				M2 =  -U*m1 		 +   m3            ;  /* -U*m1 ! */
				M3 =  -V*m1 		 		+ 	m4     ;  /* -V*m1 ! */
				M4 = S41*m1 + S42*m2 +   k2 +   k3 + kk;
				M5 = S51*m1 + S52*m2 +   k2 +   k3 + kk;
					/* backward */
				kk = K_1 * w5;
				k2 = cc * w3;
				k3 = dd * w4;
				W1 = S11*w1 + S12*w2 +   k2 +   k3 + kk;
				W2 =  -U*w1 		 +   w3 	       ;  /* -U*m1 ! */
				W3 =  -V*w1 				+   w4     ;  /* -V*m1 ! */
				W4 = S41*w1 + S42*w2 +   k2 +   k3 + kk;
				W5 = S51*w1 + S52*w2 +   k2 +   k3 + kk;
				/* modified characteristic differences _W */
				_M1 = minmod( M1, W1 );
				_M2 = minmod( M2, W2 );
				_M3 = minmod( M3, W3 );
				_M4 = minmod( M4, W4 );
				_M5 = minmod( M5, W5 );
				_W1 = minmod( W1, M1 );
				_W2 = minmod( W2, M2 );
				_W3 = minmod( W3, M3 );
				_W4 = minmod( W4, M4 );
				_W5 = minmod( W5, M5 );
				/* inverted transformation matrix [S-1] */
				/*_S21 =  w_hh;  _S31 =  u_hh; _S41 =  v_hh; */
									/*_S34 =  u_ee; _S44 =  w_ee; */
				_S24 = w_ee + ll; _S25 = w_ee - ll;
				_S54 = _S55 = nn + _1_2_K_1;
				_S54 += mm; _S55 -= mm;
				/* parameters' vector at the nearer interface */
				k1 = _M1 * C1 + _W1 * C2;
				k2 = _M2 * C1 + _W2 * C2;
				k3 = _M3 * C1 + _W3 * C2;
				k4 = _M4 * C1 + _W4 * C2;
				k5 = _M5 * C1 + _W5 * C2;
				kk = k4 + k5;
				zU1[_i][_j][ k] = u1 +   hh * k1 + 						 	ee  * kk;
				zU4[_i][_j][ k] = u4 + w_hh * k1 + 						   _S24 * k4 + _S25 * k5;
				zU2[_i][_j][ k] = u2 + u_hh * k1 + 		  k2 			 + u_ee * kk;
				zU3[_i][_j][ k] = u3 + v_hh * k1 + 					  k3 + w_ee * kk;
				zU5[_i][_j][ k] = u5 +   gg * k1 +    U * k2 +    V * k3 + _S54 * k4 + _S55 * k5;
				/* ... farther interface */
				k1 = _M1 * C2 + _W1 * C1;
				k2 = _M2 * C2 + _W2 * C1;
				k3 = _M3 * C2 + _W3 * C1;
				k4 = _M4 * C2 + _W4 * C1;
				k5 = _M5 * C2 + _W5 * C1;
				kk =  k4 + k5;
				U1z[_i][_j][_k] = u1 -   hh * k1 - 						     ee * kk;
				U4z[_i][_j][_k] = u4 - w_hh * k1 - 						   _S24 * k4 - _S25 * k5;
				U2z[_i][_j][_k] = u2 - u_hh * k1 - 		  k2 			 - u_ee * kk;
				U3z[_i][_j][_k] = u3 - v_hh * k1 - 					  k3 - w_ee * kk;
				U5z[_i][_j][_k] = u5 -   gg * k1 -    U * k2 -    V * k3 - _S54 * k4 - _S55 * k5;
			} /* end for */
		} /* end for */
	} /* end for */

} /* end Reconstruction() */


/***********************
* BOUNCONDONINTERFACES *  complete the "outer" parameters' values at the domain boundary
***********************/
void BounCondOnInterfaces(
	int myid, // identifier of _this_ process
	int numprocs )
{
int next, previous; // processes next/previous to this
MPI_Status status;

	//--- Communicate via MPI to complete outer parameters at cell x-bondaries

	//-- sending to next processes

	// next process's ID
	next = myid+1;

		// root process
	if(0 == myid) {
		// sending to the next
		for (i = 0, j = 0; j < HIG; j++) {
			for (k = 0; k < DEP; k++, i += 5) {
				Buf[i  ] = xU1[LEN][j][k];
				Buf[i+1] = xU2[LEN][j][k];
				Buf[i+2] = xU3[LEN][j][k];
				Buf[i+3] = xU4[LEN][j][k];
				Buf[i+4] = xU5[LEN][j][k];
			}
		}
		MPI_Send(Buf, BufCountF, MPI_FLOAT, next, 99, MPI_COMM_WORLD);
		// previous doesn't exist
/* 	MPI_Recv(Buf, BufCountF, MPI_FLOAT, MPI_ANY_SOURCE, 99, MPI_COMM_WORLD,
					&status);
		for (i = 0, j = 0; j < HIG; j++) {
			for (k = 0; k < DEP; k++, i += 5) {
				xU1[0][j][k] = Buf[i  ];
				xU2[0][j][k] = Buf[i+1];
				xU3[0][j][k] = Buf[i+2];
				xU4[0][j][k] = Buf[i+3];
				xU5[0][j][k] = Buf[i+4];
			}
		}
*/
	}

	// other processes
	else {
		// receiving from the previous, everyone
		MPI_Recv(Buf, BufCountF, MPI_FLOAT, MPI_ANY_SOURCE, 99, MPI_COMM_WORLD,
					&status);
		for (i = 0, j = 0; j < HIG; j++) {
			for (k = 0; k < DEP; k++, i += 5) {
				xU1[0][j][k] = Buf[i  ];
				xU2[0][j][k] = Buf[i+1];
				xU3[0][j][k] = Buf[i+2];
				xU4[0][j][k] = Buf[i+3];
				xU5[0][j][k] = Buf[i+4];
			}
		}
		// sending to the next; if not the last!
		if(myid != numprocs-1) {
			for (i = 0, j = 0; j < HIG; j++) {
				for (k = 0; k < DEP; k++, i += 5) {
					Buf[i  ] = xU1[LEN][j][k];
					Buf[i+1] = xU2[LEN][j][k];
					Buf[i+2] = xU3[LEN][j][k];
					Buf[i+3] = xU4[LEN][j][k];
					Buf[i+4] = xU5[LEN][j][k];
				}
			}
			MPI_Send(Buf, BufCountF, MPI_FLOAT, next, 99, MPI_COMM_WORLD);
		} // end if(myid != numprocs-1)
	}

	//-- sending to previous processes

	// previous process's ID
	previous = myid-1;

	// root process
	if(0 == myid) {
		// previous doesn't exist
/* 	for (i = 0, j = 0; j < HIG; j++) {
			for (k = 0; k < DEP; k++, i += 5) {
				Buf[i  ] = U1x[0][j][k];
				Buf[i+1] = U2x[0][j][k];
				Buf[i+2] = U3x[0][j][k];
				Buf[i+3] = U4x[0][j][k];
				Buf[i+4] = U5x[0][j][k];
			}
		}
		MPI_Send(Buf, BufCountF, MPI_FLOAT, previous, 99, MPI_COMM_WORLD);
*/
		// receiving from the next
		MPI_Recv(Buf, BufCountF, MPI_FLOAT, MPI_ANY_SOURCE, 99, MPI_COMM_WORLD,
					&status);
		for (i = 0, j = 0; j < HIG; j++) {
			for (k = 0; k < DEP; k++, i += 5) {
				U1x[LEN][j][k] = Buf[i  ];
				U2x[LEN][j][k] = Buf[i+1];
				U3x[LEN][j][k] = Buf[i+2];
				U4x[LEN][j][k] = Buf[i+3];
				U5x[LEN][j][k] = Buf[i+4];
			}
		}
	}

	// other processes
	else {
		// receiving from the next; if not the last process!
		if(myid != numprocs-1) {
			MPI_Recv(Buf, BufCountF, MPI_FLOAT, MPI_ANY_SOURCE, 99, MPI_COMM_WORLD,
						&status);
			for (i = 0, j = 0; j < HIG; j++) {
				for (k = 0; k < DEP; k++, i += 5) {
					U1x[LEN][j][k] = Buf[i  ];
					U2x[LEN][j][k] = Buf[i+1];
					U3x[LEN][j][k] = Buf[i+2];
					U4x[LEN][j][k] = Buf[i+3];
					U5x[LEN][j][k] = Buf[i+4];
				}
			}
		} // wnd if(myid != numprocs-1)
		// sending to previous - everyone
		for (i = 0, j = 0; j < HIG; j++) {
			for (k = 0; k < DEP; k++, i += 5) {
				Buf[i  ] = U1x[0][j][k];
				Buf[i+1] = U2x[0][j][k];
				Buf[i+2] = U3x[0][j][k];
				Buf[i+3] = U4x[0][j][k];
				Buf[i+4] = U5x[0][j][k];
			}
		}
		MPI_Send(Buf, BufCountF, MPI_FLOAT, previous, 99, MPI_COMM_WORLD);
	}

	// syncronization
	MPI_Barrier(MPI_COMM_WORLD);


	/*--- x ---*/
	/*- left side - JET inside a circle */
	if(0 == myid) { /* root process */
		P = 100000.;
		R = 1.0;
		for (j = 0; j < HIG; j++) {
			for( k = 0; k < DEP; k++ ) {
				ly = deltaY*((float)j-(float)(j_c-1));
				lz = deltaZ*((float)k-(float)(k_c-1));
				L = sqrt(ly*ly + lz*lz);
				if (L <= Radius){
					U = 200.;
					V = Vd;
					W = Wd;
				}
				else {
					U = 50.;
					V = W = 0.;
				}
				xU1[0][j][k] = R;
				xU2[0][j][k] = R * U;
				xU3[0][j][k] = R * V;
				xU4[0][j][k] = R * W;
				xU5[0][j][k] = P / K_1 + 0.5 * R * ( U * U + V * V + W * W);
			} /* end for() */
		} /* end for() */
	} // end if(0 == myid) {

	if(myid == numprocs-1) { /* last process */
		/* right side - OUTFLOW */
		P = 100000.;
		for (j = 0; j < HIG; j++) {
			for (k = 0; k < DEP; k++) {
				U1x[LEN][j][k] = R = xU1[LEN][j][k];
				U2x[LEN][j][k] = U = xU2[LEN][j][k];
				U3x[LEN][j][k] = V = xU3[LEN][j][k];
				U4x[LEN][j][k] = W = xU4[LEN][j][k];
				U5x[LEN][j][k] = P / K_1 + 0.5 * (U*U + V*V + W*W ) / R;
			} /* end for() */
		} /* end for() */
	} // end if(myid == numprocs-1)

	/*--- z SLIP ---*/
	for (i = 0; i < LEN; i++) {
		for (j = 0; j < HIG; j++) {
			/* back side - PERIODIC */
			zU1[i][j][0] =   zU1[i][j][DEP];
			zU2[i][j][0] =   zU2[i][j][DEP];
			zU3[i][j][0] =   zU3[i][j][DEP];
			zU4[i][j][0] = - zU4[i][j][DEP];
			zU5[i][j][0] =   zU5[i][j][DEP];
			/* front side - PERIODIC */
			U1z[i][j][DEP] =   U1z[i][j][0];
			U2z[i][j][DEP] =   U2z[i][j][0];
			U3z[i][j][DEP] =   U3z[i][j][0];
			U4z[i][j][DEP] = - U4z[i][j][0];
			U5z[i][j][DEP] =   U5z[i][j][0];
		} /* end for */
	} /* end for */

	/*--- y SLIP ---*/
	for (i = 0; i < LEN; i++) {
		for (k = 0; k < DEP; k++) {
			/* bottom side - SLIP */
			yU1[i][0][k] =   U1y[i][0][k];
			yU2[i][0][k] =   U2y[i][0][k];
			yU3[i][0][k] = - U3y[i][0][k];
			yU4[i][0][k] =   U4y[i][0][k];
			yU5[i][0][k] =   U5y[i][0][k];
			/* top side - SLIP */
			U1y[i][HIG][k] =   yU1[i][HIG][k];
			U2y[i][HIG][k] =   yU2[i][HIG][k];
			U3y[i][HIG][k] = - yU3[i][HIG][k];
			U4y[i][HIG][k] =   yU4[i][HIG][k];
			U5y[i][HIG][k] =   yU5[i][HIG][k];
		} /* end for */
	}  /* end for */
} /* end BounCondOnInterfaces( void ) */


/***********
*  FLUXES  *   fluxes at the interfaces
***********/
void Fluxes( void )
{
unsigned i, j, k, i_, j_, k_, i__, j__, k__;
register float RU; /* convective normal mass flux */

	 /*--- X-fluxes ---*/
for (i = 0, i_ = 1; i < LENN; i++, i_++) {
	for (j = 0, j_ = 1, j__ = 2; j < HIG; j++, j_++, j__++) {
		for (k = 0, k_ = 1, k__ = 2; k < DEP; k++, k_++, k__++) {
		/*--- Inviscid fluxes ---*/
		/* left parameters */
		_U =   xU2[i][j][k] / (_R = xU1[i][j][k]);
		_V =   xU3[i][j][k] /  _R;
		_W =   xU4[i][j][k] /  _R;
		_P = ( xU5[i][j][k] - 0.5 * _R * ( _U*_U + _V*_V + _W*_W ) ) * K_1;
		_C  = sqrt(K * _P / _R);
		_jo = _K_1 * _P; /*_ P - _C * _C * _R */
		_jp = _U + ( RU = _P / ( _R * _C ) );
		_jm = _U -   RU;
		/* right parameters */
		U_ =   U2x[i][j][k] / (R_ = U1x[i][j][k]);
		V_ =   U3x[i][j][k] /  R_;
		W_ =   U4x[i][j][k] /  R_;
		P_ = ( U5x[i][j][k] - 0.5 * R_ * ( U_*U_ + V_*V_ + W_*W_ ) ) * K_1;
		C_  = sqrt(K * P_ / R_);
		jo_ = _K_1 * P_; /* P_ - C_ * C_ * R_ */
		jp_ = U_ + ( RU = P_ / ( R_ * C_ ) );
		jm_ = U_ -   RU;
		/* linearized procedure */
		U = 0.5 * ( _U + U_ );
		C = 0.5 * ( _C + C_ );
		C_p = U + C; C_m = U - C; C_o = U;
		/*labelx:*/
		/* J_p and al_p */
		if( C_p > 0. ) { al_p = K_1_2 * ( _jm - _jp ) / _jo; J_p  = _jp; }
		else           { al_p = K_1_2 * ( jm_ - jp_ ) / jo_; J_p  = jp_; }
		/* J_m and al_m */
		if( C_m > 0. ) { al_m = K_1_2 * ( _jp - _jm ) / _jo; J_m  = _jm; }
		else           { al_m = K_1_2 * ( jp_ - jm_ ) / jo_; J_m  = jm_; }
		/* J_o and al_o */
		if( C_o > 0. ) { al_o = _05K * ( _jp - _jm ); al_o = - al_o * al_o; J_o  = _jo; }
		else           { al_o = _05K * ( jp_ - jm_ ); al_o = - al_o * al_o; J_o  = jo_; }
		/* parameters at the interface */
		P = ( J_p - J_m ) / ( al_p - al_m );
		U = J_p - al_p * P;
		R = ( J_o - P ) / al_o;
		/* check signs of the Ch-velocities
		C = sqrt( K * P / R );
		C_p_ = U + C; C_m_ = U - C; C_o_ = U;
		if( C_p_*C_p < 0 || C_m_*C_m < 0 || C_o_*C_o < 0 ) {
			 C_p = C_p_; C_m = C_m_; C_o = C_o_;
			 goto labelx;
		}
		*/
		/* tangential velocities */
		if( U > 0 ) { V = _V; W = _W; }
		else {        V = V_; W = W_; }
		/*--- Viscous fluxes ---*/
		/*-- forward cells --*/
		/* forward */
		U_ =    U2_[i_ ][j_ ][k_ ] / (R_ = U1_[i_ ][j_ ][k_ ]);
		V_ =    U3_[i_ ][j_ ][k_ ] /  R_;
		W_ =    U4_[i_ ][j_ ][k_ ] /  R_;
		P_ = (  U5_[i_ ][j_ ][k_ ] - 0.5 * R_ * ( U_ * U_ + V_ * V_ + W_ * W_ ) ) * K_1;
		T_ = P_ / ( R_VOZD * R_ );
		/* forward upper */
		Uj =    U2_[i_ ][j__][k_ ] * (RU = 1./U1_[i_ ][j__][k_ ]);
		Vj =    U3_[i_ ][j__][k_ ] *  RU;
		Wj =    U4_[i_ ][j__][k_ ] *  RU;
		/* forward lower */
		jU =    U2_[i_ ][j  ][k_ ] * (RU = 1./U1_[i_ ][j  ][k_ ]);
		jV =    U3_[i_ ][j  ][k_ ] *  RU;
		jW =    U4_[i_ ][j  ][k_ ] *  RU;
		/* forward nearer */
		Uk =    U2_[i_ ][j_ ][k__] * (RU = 1./U1_[i_ ][j_ ][k__]);
		Vk =    U3_[i_ ][j_ ][k__] *  RU;
		Wk =    U4_[i_ ][j_ ][k__] *  RU;
		/* forward farther */
		kU =    U2_[i_ ][j_ ][k  ] * (RU = 1./U1_[i_ ][j_ ][k  ]);
		kV =    U3_[i_ ][j_ ][k  ] *  RU;
		kW =    U4_[i_ ][j_ ][k  ] *  RU;
		/*-- backward cells-- */
		/* backward */
		_U =    U2_[i  ][j_ ][k_ ] / (_R = U1_[i  ][j_ ][k_ ]);
		_V =    U3_[i  ][j_ ][k_ ] /  _R;
		_W =    U4_[i  ][j_ ][k_ ] /  _R;
		_P = (  U5_[i  ][j_ ][k_ ] - 0.5 * _R * ( _U * _U + _V * _V + _W * _W ) ) * K_1;
		_T = _P / ( R_VOZD * _R );
		/* backward upper */
		uj =    U2_[i  ][j__][k_ ] * (RU = 1./U1_[i  ][j__][k_ ]);
		vj =    U3_[i  ][j__][k_ ] *  RU;
		wj =    U4_[i  ][j__][k_ ] *  RU;
		/* backward lower */
		ju =    U2_[i  ][j  ][k_ ] * (RU = 1./U1_[i  ][j  ][k_ ]);
		jv =    U3_[i  ][j  ][k_ ] *  RU;
		jw =    U4_[i  ][j  ][k_ ] *  RU;
		/* backward nearer */
		uk =    U2_[i  ][j_ ][k__] * (RU = 1./U1_[i  ][j_ ][k__]);
		vk =    U3_[i  ][j_ ][k__] *  RU;
		wk =    U4_[i  ][j_ ][k__] *  RU;
		/* backward farther */
		ku =    U2_[i  ][j_ ][k  ] * (RU = 1./U1_[i  ][j_ ][k  ]);
		kv =    U3_[i  ][j_ ][k  ] *  RU;
		kw =    U4_[i  ][j_ ][k  ] *  RU;
		/* derivatives of velocities */
		/* x */
		du_dx = (U_ - _U) * _deltaX;
		dv_dx = (V_ - _V) * _deltaX;
		dw_dx = (W_ - _W) * _deltaX;
		/* y */
		du_dy = ( Uj + uj - jU - ju ) * _4deltaY;
		dv_dy = ( Vj + vj - jV - jv ) * _4deltaY;
		dw_dy = ( Wj + wj - jW - jw ) * _4deltaY;
		/* z */
		du_dz = ( Uk + uk - kU - ku ) * _4deltaZ;
		dv_dz = ( Vk + vk - kV - kv ) * _4deltaZ;
		dw_dz = ( Wk + wk - kW - kw ) * _4deltaZ;
		/* mean velocities and density */
		rr = 0.5 * ( R_ + _R );
		uu = 0.5 * ( U_ + _U );
		vv = 0.5 * ( V_ + _V );
		ww = 0.5 * ( W_ + _W );
		/* stresses */
		S12 = du_dy + dv_dx;
		S13 = du_dz + dw_dx;
		S23 = dv_dz + dw_dy;
		_S_ = sqrt( 2. * ( du_dx * du_dx + dv_dy * dv_dy + dw_dz * dw_dz )
			  +      ( S12   * S12   + S13   * S13   + S23   * S23 ) );
		mu_E = mu_L + ( mu_T = rr * CsDD * _S_ ); /* mu_T = r * Cs * delta * delta * | S |  */
		sigma_xx = 0.66666667 * mu_E * ( du_dx + du_dx - dv_dy - dw_dz );
		sigma_xy = mu_E * ( du_dy + dv_dx );
		sigma_xz = mu_E * ( du_dz + dw_dx );
		/* heat flux */
		lambda_E = lambda_L + mu_T * cp_Pr_T;
		q_x = lambda_E * ( _T - T_ ) * _deltaX;
		/* summary X-fluxes evaluation */
		Fx1[i][j][k] = RU = R * U;
		Fx2[i][j][k] = RU * U + P - sigma_xx;
		Fx3[i][j][k] = RU * V     - sigma_xy;
		Fx4[i][j][k] = RU * W	  - sigma_xz;
		Fx5[i][j][k] = U * ( K_K_1 * P + 0.5 * R * ( U * U + V * V + W * W ) )
				- uu * sigma_xx - vv * sigma_xy - ww * sigma_xz + q_x;
		}
	}
}

	 /*--- Y-fluxes ---*/
for( i = 0, i_ = 1, i__ = 2; i < LEN; i++, i_++, i__++ ) {
	for( j = 0, j_ = 1; j < HIGG; j++, j_++ ) {
		for( k = 0, k_ = 1, k__ = 2; k < DEP; k++, k_++, k__++ ) {
		/*--- Inviscid fluxes ---*/
		/* lower */
		_U =   yU2[i][j][k] / (_R = yU1[i][j][k]);
		_V =   yU3[i][j][k] /  _R;
		_W =   yU4[i][j][k] /  _R;
		_P = ( yU5[i][j][k] - 0.5 * _R * ( _U*_U + _V*_V + _W*_W ) ) * K_1;
		_C  = sqrt( K * _P / _R );
		_jo = _K_1 * _P; /* _P - _C * _C * _R */
		_jp = _V + ( RU = _P / ( _R * _C ) );
		_jm = _V -   RU;
		/* upper */
		U_ =   U2y[i][j][k] / (R_ = U1y[i][j][k]);
		V_ =   U3y[i][j][k] /  R_;
		W_ =   U4y[i][j][k] /  R_;
		P_ = ( U5y[i][j][k] - 0.5 * R_ * ( U_*U_ + V_*V_ + W_*W_ ) ) * K_1;
		C_  = sqrt( K * P_ / R_ );
		jo_ = _K_1 * P_; /* P_ - C_ * C_ * R_ */
		jp_ = V_ + ( RU = P_ / ( R_ * C_ ) );
		jm_ = V_ -   RU;
		/* linearized procedure */
		V = 0.5 * ( _V + V_ );
		C = 0.5 * ( _C + C_ );
		C_p = V + C; C_m = V - C; C_o = V;
		/*labely:*/
		/* Jp and al_p */
		if( C_p > 0. ) { al_p = K_1_2 * ( _jm - _jp ) / _jo; J_p  = _jp; }
		else           { al_p = K_1_2 * ( jm_ - jp_ ) / jo_; J_p  = jp_; }
		/* Jm and al_m */
		if( C_m > 0. ) { al_m = K_1_2 * ( _jp - _jm ) / _jo; J_m  = _jm; }
		else           { al_m = K_1_2 * ( jp_ - jm_ ) / jo_; J_m  = jm_; }
		/* Jo and al_o */
		if( C_o > 0. ) { al_o = _05K * ( _jp - _jm ); al_o = - al_o * al_o; J_o  = _jo; }
		else           { al_o = _05K * ( jp_ - jm_ ); al_o = - al_o * al_o; J_o  = jo_; }
		/* parameters at the interface */
		P = ( J_p - J_m ) / ( al_p - al_m );
		V = J_p - al_p * P;
		R = ( J_o - P ) / al_o;
		/* check signs of the Ch-velocities
		C = sqrt( K * P / R );
		C_p_ = V + C; C_m_ = V - C; C_o_ = V;
		if( C_p_*C_p < 0 || C_m_*C_m < 0 || C_o_*C_o < 0 ) {
			C_p = C_p_; C_m = C_m_; C_o = C_o_;
			goto labely;
		}
		*/
		/* tangential velocities */
		if( V > 0 ) { U = _U; W = _W; }
		else {        U = U_; W = W_; }
		/*--- Viscous fluxes ---*/
		/*-- upper cells --*/
		/* upper */
		U_ =    U2_[i_ ][j_ ][k_ ] / (R_ = U1_[i_ ][j_ ][k_ ]);
		V_ =    U3_[i_ ][j_ ][k_ ] / R_;
		W_ =    U4_[i_ ][j_ ][k_ ] / R_;
		P_ = (  U5_[i_ ][j_ ][k_ ] - 0.5 * R_ * ( U_*U_ + V_*V_ + W_*W_ ) ) * K_1;
		T_ = P_ / ( R_VOZD * R_ );
		/* upper nearer */
		Uk =    U2_[i_ ][j_ ][k__] * (RU = 1./U1_[i_ ][j_ ][k__]);
		Vk =    U3_[i_ ][j_ ][k__] *  RU;
		Wk =    U4_[i_ ][j_ ][k__] *  RU;
		/* upper farther */
		kU =    U2_[i_ ][j_ ][k  ] * (RU = 1./U1_[i_ ][j_ ][k  ]);
		kV =    U3_[i_ ][j_ ][k  ] *  RU;
		kW =    U4_[i_ ][j_ ][k  ] *  RU;
		/* upper forward */
		Ui =    U2_[i__][j_ ][k_ ] * (RU = 1./U1_[i__][j_ ][k_ ]);
		Vi =    U3_[i__][j_ ][k_ ] *  RU;
		Wi =    U4_[i__][j_ ][k_ ] *  RU;
		/* upper backward */
		iU =    U2_[i  ][j_ ][k_ ] * (RU = 1./U1_[i  ][j_ ][k_ ]);
		iV =    U3_[i  ][j_ ][k_ ] *  RU;
		iW =    U4_[i  ][j_ ][k_ ] *  RU;
		/*-- lower cells--*/
		/* lower */
		_U =    U2_[i_ ][j  ][k_ ] / (_R = U1_[i_ ][j  ][k_ ]);
		_V =    U3_[i_ ][j  ][k_ ] /  _R;
		_W =    U4_[i_ ][j  ][k_ ] /  _R;
		_P = (  U5_[i_ ][j  ][k_ ] - 0.5 * _R * ( _U*_U + _V*_V + _W*_W ) ) * K_1;
		_T = _P / ( R_VOZD * _R );
		/* lower nearer */
		uk =    U2_[i_ ][j  ][k__] * (RU = 1./U1_[i_ ][j  ][k__]);
		vk =    U3_[i_ ][j  ][k__] *  RU;
		wk =    U4_[i_ ][j  ][k__] *  RU;
		/* lower farther */
		ku =    U2_[i_ ][j  ][k  ] * (RU = 1./U1_[i_ ][j  ][k  ]);
		kv =    U3_[i_ ][j  ][k  ] *  RU;
		kw =    U4_[i_ ][j  ][k  ] *  RU;
		/* lower forward */
		ui =    U2_[i__][j  ][k_ ] * (RU = 1./U1_[i__][j  ][k_ ]);
		vi =    U3_[i__][j  ][k_ ] *  RU;
		wi =    U4_[i__][j  ][k_ ] *  RU;
		/* lower backward */
		iu =    U2_[i  ][j  ][k_ ] * (RU = 1./U1_[i  ][j  ][k_ ]);
		iv =    U3_[i  ][j  ][k_ ] *  RU;
		iw =    U4_[i  ][j  ][k_ ] *  RU;
		/* derivatives of velocities */
		/* x */
		du_dx = ( Ui + ui - iU - iu ) * _4deltaX;
		dv_dx = ( Vi + vi - iV - iv ) * _4deltaX;
		dw_dx = ( Wi + wi - iW - iw ) * _4deltaX;
		/* y */
		du_dy = ( U_ - _U ) * _deltaY;
		dv_dy = ( V_ - _V ) * _deltaY;
		dw_dy = ( W_ - _W ) * _deltaY;
		/* z */
		du_dz = ( Uk + uk - kU - ku ) * _4deltaZ;
		dv_dz = ( Vk + vk - kV - kv ) * _4deltaZ;
		dw_dz = ( Wk + wk - kW - kw ) * _4deltaZ;
		/* mean velocities and density */
		rr = 0.5 * (R_ + _R);
		uu = 0.5 * (U_ + _U);
		vv = 0.5 * (V_ + _V);
		ww = 0.5 * (W_ + _W);
		/* stresses */
		S12 = du_dy + dv_dx;
		S13 = du_dz + dw_dx;
		S23 = dv_dz + dw_dy;
		_S_ = sqrt( 2. * ( du_dx*du_dx + dv_dy * dv_dy + dw_dz * dw_dz )
					 +      ( S12  * S12  + S13   * S13   + S23   * S23 ) );
		mu_E = mu_L + ( mu_T = rr * CsDD * _S_ );  /* mu_T = r * Cs * delta * delta * | S |  */
		sigma_yx = mu_E * ( dv_dx + du_dy );
		sigma_yy = 0.66666667 * mu_E * ( dv_dy + dv_dy - du_dx - dw_dz );
		sigma_yz = mu_E * ( dv_dz + dw_dy );
		/* heat flux */
		lambda_E = lambda_L + mu_T * cp_Pr_T;
		q_y = lambda_E * (_T - T_) * _deltaY;
		/* summary Y-fluxes */
		Fy1[i][j][k] = RU = R * V;
		Fy2[i][j][k] = RU * U     - sigma_yx;
		Fy3[i][j][k] = RU * V + P - sigma_yy;
		Fy4[i][j][k] = RU * W     - sigma_yz;
		Fy5[i][j][k] = V * ( K_K_1 * P + 0.5 * R * ( U * U + V * V + W * W ) )
				  - uu * sigma_yx - vv * sigma_yy - ww * sigma_yz + q_y;
		 }
	}
}

	 /*--- Z-fluxes ---*/
for (i = 0, i_ = 1, i__ = 2; i < LEN; i++, i_++, i__++) {
	for (j = 0, j_ = 1, j__ = 2; j < HIG; j++, j_++, j__++) {
		 for (k = 0, k_ = 1; k < DEPP; k++, k_++) {
		/*--- Inviscid fluxes ---*/
		/* farther */
		_U =   zU2[i][j][k] / (_R = zU1[i][j][k]);
		_V =   zU3[i][j][k] /  _R;
		_W =   zU4[i][j][k] /  _R;
		_P = ( zU5[i][j][k] - 0.5 * _R * ( _U*_U + _V*_V + _W*_W ) ) * K_1;
		_C  = sqrt( K * _P / _R );
		_jo = _K_1 * _P; /* _P - _C * _C * _R; */
		_jp = _W + ( RU = _P / ( _R * _C ) );
		_jm = _W -   RU;
		/* nearer */
		U_ =   U2z[i][j][k] / (R_ = U1z[i][j][k] );
		V_ =   U3z[i][j][k] /  R_;
		W_ =   U4z[i][j][k] /  R_;
		P_ = ( U5z[i][j][k] - 0.5 * R_ * ( U_*U_ + V_*V_ + W_*W_ ) ) * K_1;
		C_  = sqrt( K * P_ / R_ );
		jo_ = _K_1 * P_; /* P_ - C_ * C_ * R_; */
		jp_ = W_ + ( RU = P_ / ( R_ * C_ ) );
		jm_ = W_ -   RU;
		/* linearized procedure */
		W = 0.5 * ( _W + W_ );
		C = 0.5 * ( _C + C_ );
		C_p = W + C; C_m = W - C; C_o = W;
		/*labelz:*/
		/* Jp and al_p */
		if( C_p > 0. ) { al_p = K_1_2 * ( _jm - _jp ) / _jo; J_p  = _jp; }
		else 	         { al_p = K_1_2 * ( jm_ - jp_ ) / jo_; J_p  = jp_; }
		/* Jm and al_m */
		if( C_m > 0. ) { al_m = K_1_2 * ( _jp - _jm ) / _jo; J_m  = _jm; }
		else           { al_m = K_1_2 * ( jp_ - jm_ ) / jo_; J_m  = jm_; }
		/* Jo and al_o */
		if( C_o > 0. ) { al_o = _05K * ( _jp - _jm ); al_o = - al_o * al_o; J_o  = _jo; }
		else           { al_o = _05K * ( jp_ - jm_ ); al_o = - al_o * al_o; J_o  = jo_; }
		/* parameters at the interface */
		P = ( J_p - J_m ) / ( al_p - al_m );
		W = J_p - al_p * P;
		R = ( J_o - P ) / al_o;
		/* check signs of the Ch-velocities
		C = sqrt( K * P / R );
		C_p_ = W + C; C_m_ = W - C; C_o_ = W;
		if( C_p_*C_p < 0 || C_m_*C_m < 0 || C_o_*C_o < 0 ) {
			 C_p = C_p_; C_m = C_m_; C_o = C_o_;
			 goto labelz;
		}
		*/
		/* tangential velocities */
		if( W > 0 ) { U = _U; V = _V; }
		else {        U = U_; V = V_; }
		/*--- Viscous fluxes ---*/
		/*-- nearer cells-- */
		/* nearer */
		U_ =    U2_[i_ ][j_ ][k_ ] / (R_ = U1_[i_ ][j_ ][k_ ]);
		V_ =    U3_[i_ ][j_ ][k_ ] /  R_;
		W_ =    U4_[i_ ][j_ ][k_ ] /  R_;
		P_ = (  U5_[i_ ][j_ ][k_ ] - 0.5 * R_ * ( U_ * U_ + V_ * V_ + W_ * W_ ) ) * K_1;
		T_ = P_ / ( R_VOZD * R_ );
		/* nearer forward */
		Ui =    U2_[i__][j_ ][k_ ] * (RU = 1./U1_[i__][j_ ][k_ ]);
		Vi =    U3_[i__][j_ ][k_ ] *  RU;
		Wi =    U4_[i__][j_ ][k_ ] *  RU;
		/* nearer backward */
		iU =    U2_[i  ][j_ ][k_ ] * (RU = 1./U1_[i  ][j_ ][k_ ]);
		iV =    U3_[i  ][j_ ][k_ ] *  RU;
		iW =    U4_[i  ][j_ ][k_ ] *  RU;
		/* nearer upper */
		Uj =    U2_[i_ ][j__][k_ ] * (RU = 1./U1_[i_ ][j__][k_ ]);
		Vj =    U3_[i_ ][j__][k_ ] * RU;
		Wj =    U4_[i_ ][j__][k_ ] * RU;
		/* nearer lower */
		jU =    U2_[i_ ][j  ][k_ ] * (RU= 1./U1_[i_ ][j  ][k_ ]);
		jV =    U3_[i_ ][j  ][k_ ] *  RU;
		jW =    U4_[i_ ][j  ][k_ ] *  RU;
		/*-- farther cells--*/
		/* farther */
		_U =    U2_[i_ ][j_ ][k  ] / (_R = U1_[i_ ][j_ ][k  ]);
		_V =    U3_[i_ ][j_ ][k  ] /  _R;
		_W =    U4_[i_ ][j_ ][k  ] /  _R;
		_P = (  U5_[i_ ][j_ ][k  ] - 0.5 * _R * ( _U * _U + _V * _V + _W * _W ) ) * K_1;
		_T = _P / ( R_VOZD * _R );
		/* farther forward */
		ui =    U2_[i__][j_ ][k  ] * (RU = 1./U1_[i__][j_ ][k  ]);
		vi =    U3_[i__][j_ ][k  ] *  RU;
		wi =    U4_[i__][j_ ][k  ] *  RU;
		/* farther backward */
		iu =    U2_[i  ][j_ ][k  ] * (RU = 1./U1_[i  ][j_ ][k  ]);
		iv =    U3_[i  ][j_ ][k  ] *  RU;
		iw =    U4_[i  ][j_ ][k  ] *  RU;
		/* farther upper */
		uj =    U2_[i_ ][j__][k  ] * (RU = 1./U1_[i_ ][j__][k  ]);
		vj =    U3_[i_ ][j__][k  ] *  RU;
		wj =    U4_[i_ ][j__][k  ] *  RU;
		/* farther lower */
		ju =    U2_[i_ ][j  ][k  ] * (RU = 1./U1_[i_ ][j  ][k  ]);
		jv =    U3_[i_ ][j  ][k  ] *  RU;
		jw =    U4_[i_ ][j  ][k  ] *  RU;
		/* derivatives of velocities */
		/* x */
		du_dx = ( Ui + ui - iU - iu ) * _4deltaX;
		dv_dx = ( Vi + vi - iV - iv ) * _4deltaX;
		dw_dx = ( Wi + wi - iW - iw ) * _4deltaX;
		/* y */
		du_dy = ( Uj + uj - jU - ju ) * _4deltaY;
		dv_dy = ( Vj + vj - jV - jv ) * _4deltaY;
		dw_dy = ( Wj + wj - jW - jw ) * _4deltaY;
		/* z */
		du_dz = (U_ - _U) * _deltaZ;
		dv_dz = (V_ - _V) * _deltaZ;
		dw_dz = (W_ - _W) * _deltaZ;
		/* mean velocities and density */
		rr = 0.5 * (R_ + _R);
		uu = 0.5 * (U_ + _U);
		vv = 0.5 * (V_ + _V);
		ww = 0.5 * (W_ + _W);
		/* stresses */
		S12 = du_dy + dv_dx;
		S13 = du_dz + dw_dx;
		S23 = dv_dz + dw_dy;
		_S_ = sqrt( 2. * ( du_dx * du_dx + dv_dy * dv_dy + dw_dz * dw_dz )
			  +      ( S12   * S12   + S13   * S13   + S23   * S23 ) );
		mu_E = mu_L + ( mu_T = rr * CsDD * _S_ ); /* mu_T = r * Cs * delta * delta * | S |  */
		sigma_zx = mu_E * ( dw_dx + du_dz );
		sigma_zy = mu_E * ( dw_dy + dv_dz );
		sigma_zz = 0.66666667 * mu_E * ( dw_dz + dw_dz - du_dx - dv_dy );
		/* heat flux */
		lambda_E = lambda_L + mu_T * cp_Pr_T;
		q_z = - lambda_E * ( T_ - _T ) * _deltaZ;
		/* summary Z-fluxes evaluation */
		Fz1[i][j][k] = RU = R * W;
		Fz2[i][j][k] = RU * U 	  - sigma_zx;
		Fz3[i][j][k] = RU * V 	  - sigma_zy;
		Fz4[i][j][k] = RU * W + P - sigma_zz;
		Fz5[i][j][k] = W * ( K_K_1 * P + 0.5 * R * ( U * U + V * V + W * W ) )
						 - uu * sigma_zx - vv * sigma_zy - ww * sigma_zz + q_z;
		 }
	}
}

} /* end Fluxes() */

/**************
*  EVOLUTION  *  compute new (or intermediate) values of conservative variables
**************/
void Evolution(void)
{
unsigned i, j, k, _i, _j, _k;

	//--- 1st stage
	if(Stage == 1) {
		//
		for (i = 1, _i = 0; i < LENN; i++, _i++) {
			for (j = 1, _j = 0; j < HIGG; j++, _j++) { /* indices i-1, k-1 etc */
				for (k = 1, _k = 0; k < DEPP; k++, _k++) {
					U1p[i][j][k] = U1[i][j][k]
									+ deltaT_X * ( Fx1[_i][_j][_k] - Fx1[i][_j][_k] )
									+ deltaT_Y * ( Fy1[_i][_j][_k] - Fy1[_i][j][_k] )
									+ deltaT_Z * ( Fz1[_i][_j][_k] - Fz1[_i][_j][k] );
					U2p[i][j][k] = U2[i][j][k]
									+ deltaT_X * ( Fx2[_i][_j][_k] - Fx2[i][_j][_k] )
									+ deltaT_Y * ( Fy2[_i][_j][_k] - Fy2[_i][j][_k] )
									+ deltaT_Z * ( Fz2[_i][_j][_k] - Fz2[_i][_j][k] );
					U3p[i][j][k] = U3[i][j][k]
									+ deltaT_X * ( Fx3[_i][_j][_k] - Fx3[i][_j][_k] )
									+ deltaT_Y * ( Fy3[_i][_j][_k] - Fy3[_i][j][_k] )
									+ deltaT_Z * ( Fz3[_i][_j][_k] - Fz3[_i][_j][k] );
					U4p[i][j][k] = U4[i][j][k]
									+ deltaT_X * ( Fx4[_i][_j][_k] - Fx4[i][_j][_k] )
									+ deltaT_Y * ( Fy4[_i][_j][_k] - Fy4[_i][j][_k] )
									+ deltaT_Z * ( Fz4[_i][_j][_k] - Fz4[_i][_j][k] );
					U5p[i][j][k] = U5[i][j][k]
									+ deltaT_X * ( Fx5[_i][_j][_k] - Fx5[i][_j][_k] )
									+ deltaT_Y * ( Fy5[_i][_j][_k] - Fy5[_i][j][_k] )
									+ deltaT_Z * ( Fz5[_i][_j][_k] - Fz5[_i][_j][k] );
				 } // end for
			} // end for
		} // end for

		// next stage
		Stage = 2;
		U1_ = U1p;
		U2_ = U2p;
		U3_ = U3p;
		U4_ = U4p;
		U5_ = U5p;

	} // end if() 1st stage

	//--- 1nd stage
	else {
		//
		for ( i = 1, _i = 0; i < LENN; i++, _i++) {
			for ( j = 1, _j = 0; j < HIGG; j++, _j++) {
				for ( k = 1, _k = 0; k < DEPP; k++, _k++) {
					U1[i][j][k] = 0.5 * ( U1[i][j][k] + U1p[i][j][k]
								 + deltaT_X * ( Fx1[_i][_j][_k] - Fx1[i][_j][_k] )
								 + deltaT_Y * ( Fy1[_i][_j][_k] - Fy1[_i][j][_k] )
								 + deltaT_Z * ( Fz1[_i][_j][_k] - Fz1[_i][_j][k] )
								);
					U2[i][j][k] = 0.5 * ( U2[i][j][k] + U2p[i][j][k]
								 + deltaT_X * ( Fx2[_i][_j][_k] - Fx2[i][_j][_k] )
								 + deltaT_Y * ( Fy2[_i][_j][_k] - Fy2[_i][j][_k] )
								 + deltaT_Z * ( Fz2[_i][_j][_k] - Fz2[_i][_j][k] )
								);
					U3[i][j][k] = 0.5 * ( U3[i][j][k] + U3p[i][j][k]
								 + deltaT_X * ( Fx3[_i][_j][_k] - Fx3[i][_j][_k] )
								 + deltaT_Y * ( Fy3[_i][_j][_k] - Fy3[_i][j][_k] )
								 + deltaT_Z * ( Fz3[_i][_j][_k] - Fz3[_i][_j][k] )
								);
					U4[i][j][k] = 0.5 * ( U4[i][j][k] + U4p[i][j][k]
								 + deltaT_X * ( Fx4[_i][_j][_k] - Fx4[i][_j][_k] )
								 + deltaT_Y * ( Fy4[_i][_j][_k] - Fy4[_i][j][_k] )
								 + deltaT_Z * ( Fz4[_i][_j][_k] - Fz4[_i][_j][k] )
								);
					U5[i][j][k] = 0.5 * ( U5[i][j][k] + U5p[i][j][k]
								 + deltaT_X * ( Fx5[_i][_j][_k] - Fx5[i][_j][_k] )
								 + deltaT_Y * ( Fy5[_i][_j][_k] - Fy5[_i][j][_k] )
								 + deltaT_Z * ( Fz5[_i][_j][_k] - Fz5[_i][_j][k] )
								);
				 } // end for
			} // end for
		} // end for

		// next stage
		Stage = 1;
		U1_ = U1;
		U2_ = U2;
		U3_ = U3;
		U4_ = U4;
		U5_ = U5;

	} // end else 2nd stage

} // end Evolution()


/***********
*  PROBES  * 	any process can write to its own "probes.*" file
***********/
int Probes(int myid)
{
	//--- probes are positioned on line formed by intersection of
	//    cenral x-, z-planes
	i = LEN/2+1;
	k = j_c;

	//--- output to "probes.myid"
	// step
	fwrite(&step, sizeof(step), 1, pFprobes);
	// HIG-sized blocks of current R and RU
		//
	for (j = 1; j < HIGG; j++)
		probes[j-1] = U1[i][j][k];
	fwrite(probes, sizeof(probes[0]) * HIG, 1, pFprobes);
		//
	for (j = 1; j < HIGG; j++)
		probes[j-1] = U2[i][j][k];
	fwrite(probes, sizeof(probes[0]) * HIG, 1, pFprobes);

	//---
	return 1;
} // end Probes()


/***********
*  OUTPUT  *   Outputs flowfield to "step.myid"
***********/
void Output(int myid)
{
float buf;
char filename[30];
MPI_File fh;
MPI_Status status;

	//--- create file
	sprintf(filename, "%d.%d", step, myid);
	MPI_File_open(MPI_COMM_SELF, filename, MPI_MODE_CREATE | MPI_MODE_RDWR,
					  MPI_INFO_NULL, &fh );
	MPI_File_set_view(fh, 0, MPI_FLOAT, MPI_FLOAT, "native", MPI_INFO_NULL);

	//--- write header -> (float)???
	buf = (float)LEN;    MPI_File_write(fh, &buf, 1, MPI_FLOAT, &status);
	buf = (float)HIG;    MPI_File_write(fh, &buf, 1, MPI_FLOAT, &status);
	buf = (float)deltaX; MPI_File_write(fh, &buf, 1, MPI_FLOAT, &status);
	buf = (float)deltaY; MPI_File_write(fh, &buf, 1, MPI_FLOAT, &status);

	//--- central z-plane
	k = j_c;

	//--- write data
	for (j = 1; j < HIGG; j++) {
		for (i = 1; i < LENN; i++) {
			// vorticity plot
			dv_dx = ( U3[i+1][j][k]/U1[i+1][j][k]
					  - U3[i-1][j][k]/U1[i-1][j][k] ) * _2deltaX;
			dw_dx = ( U4[i+1][j][k]/U1[i+1][j][k]
					  - U4[i-1][j][k]/U1[i-1][j][k] ) * _2deltaX;
			du_dy = ( U2[i][j+1][k]/U1[i][j+1][k]
					  - U2[i][j-1][k]/U1[i][j-1][k] ) * _2deltaY;
			dw_dy = ( U4[i][j+1][k]/U1[i][j+1][k]
					  - U4[i][j-1][k]/U1[i][j-1][k] ) * _2deltaY;
			du_dz = ( U2[i][j][k+1]/U1[i][j][k+1]
					  - U2[i][j][k-1]/U1[i][j][k-1] ) * _2deltaZ;
			dv_dz = ( U3[i][j][k+1]/U1[i][j][k+1]
					  - U3[i][j][k-1]/U1[i][j][k-1] ) * _2deltaZ;
			U = 0.5 * ( dw_dy - dv_dz );
			V = 0.5 * ( du_dz - dw_dx );
			W = 0.5 * ( dv_dx - du_dy );
			buf = sqrt( U*U + V*V + W*W );
			MPI_File_write(fh, &buf, 1, MPI_FLOAT, &status);
		}
	}

	//--- close file
	MPI_File_close(&fh);

} // end Output()


/*************
*  FINALIZE  *  Save arrays to the binary "backup.myid" file
*************/
void Finalize(int myid)
{
float buf;
char filename[30];
MPI_File fh;
MPI_Status status;

FILE *pF;


	//--- open/create "backup.myid" file in binary mode
	sprintf(filename, "backup.%d", myid);
	MPI_File_open(MPI_COMM_SELF, filename, MPI_MODE_CREATE | MPI_MODE_RDWR,
						MPI_INFO_NULL, &fh);
	MPI_File_set_view(fh, 0, MPI_FLOAT, MPI_FLOAT, "native", MPI_INFO_NULL);

	//--- write data
	for (i = 1; i < LENN; i++) {
		for (j = 1; j < HIGG; j++) {
			MPI_File_write(fh, &U1[i][j][1], DEP, MPI_FLOAT, &status);
			MPI_File_write(fh, &U2[i][j][1], DEP, MPI_FLOAT, &status);
			MPI_File_write(fh, &U3[i][j][1], DEP, MPI_FLOAT, &status);
			MPI_File_write(fh, &U4[i][j][1], DEP, MPI_FLOAT, &status);
			MPI_File_write(fh, &U5[i][j][1], DEP, MPI_FLOAT, &status);
		}
	}
	// save current step
	buf = (float)step;
	MPI_File_write(fh, &buf, 1, MPI_FLOAT, &status);
	// save seed of pseudo-random number
	buf = (float)rand();
	MPI_File_write(fh, &buf, 1, MPI_FLOAT, &status);

	//--- close "backup.myid" file
	MPI_File_close(&fh);

	//--- close file "probes.dat" here
	fclose(pFprobes);

} // end Finalize()


/*********
*  MAIN  *   Main function
*********/
int main(int argc, char *argv[])
{
int
	numprocs, // number of processes
	myid,     // identifier of _this_ process
	namelen;

char processor_name[MPI_MAX_PROCESSOR_NAME], a;

MPI_File fh;
MPI_Status status;


	//-- MPI Initialization block
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD, &myid);

	//-- Initializations for numerical scheme
	Initialize(myid);

	/*--- time stepping ---*/
	while (1) {

		//-- Broadcast step from the process with the rank "root" to others
		MPI_Bcast(&step, 1, MPI_INT, 0, MPI_COMM_WORLD);

		// Only root process checks single-byte record in "stopfile": 'g' or 's'
		if(0 == myid) {
			MPI_File_open(MPI_COMM_SELF, "stopfile",
								MPI_MODE_CREATE | MPI_MODE_RDWR, MPI_INFO_NULL, &fh);
			MPI_File_set_view(fh, 0, MPI_CHAR, MPI_CHAR, "native", MPI_INFO_NULL);
			MPI_File_read(fh, &a, 1, MPI_CHAR, &status);
			MPI_File_close(&fh);
			//
			if (a == 's') {
				a = 'g';
				continFlag = 0;
				MPI_File_open(MPI_COMM_SELF, "stopfile",
							MPI_MODE_CREATE | MPI_MODE_RDWR,
							MPI_INFO_NULL, &fh);
							MPI_File_set_view(fh, 0, MPI_CHAR, MPI_CHAR,
							"native", MPI_INFO_NULL);
				MPI_File_write(fh, &a, 1, MPI_CHAR, &status);
				MPI_File_close(&fh);
			}
		}

		//-- Broadcast continFlag from the process with the rank "root" to others
		MPI_Bcast(&continFlag, 1, MPI_INT, 0, MPI_COMM_WORLD);

		//-- Every process may check exit conditions
		if(step == numstep || !continFlag) break;

		//-- root process prints current step and defines disturbances
		if(0 == myid) {
			//
			fprintf(stdout, "%d\n", ++step);
			//
			if(counter > 0) counter--;
			else {
				counter = (int)(Ns_min + Nst * rand() / (RAND_MAX-1.) );
				Vd = Va * (1. - 2.* rand() / (RAND_MAX-1.));
				Wd = Wa * (1. - 2.* rand() / (RAND_MAX-1.));
				fprintf(stdout, "Vd = %f, Wd = %f\n", Vd, Wd);
			}
		}

		/*=== first stage ===*/
		//--- BC in ghost cells (1)
		BounCondInGhostCells(myid, numprocs);
		//--- parameters at the cell boundaries (1)
		Reconstruction();
		//--- BC at cell boundaries at the domain boundary (1)
		BounCondOnInterfaces(myid, numprocs);
		//--- fluxes (1)
		Fluxes();
		//--- evolution in time (1)
		Evolution();

		/*=== second stage ===*/
		//--- BC in ghost cells (2)
		BounCondInGhostCells(myid, numprocs);
		//--- parameters at the cell boundaries (2)
		Reconstruction();
		//--- BC at cell boundaries at the domain boundary (2)
		BounCondOnInterfaces(myid, numprocs);
		//--- fluxes (2)
		Fluxes();
		//--- evolution in time (2)
		Evolution();

		//-- Probes output
		Probes(myid);

		//-- Take frame
		if(step%f_step == 0) Output(myid);

	} // end while(1)

	//-- Finalize: backup the solution and close probes.* files
	Finalize(myid);

	//---
	fprintf(stdout, "%d/%d process stopped!\n", myid+1, numprocs);

	//---
	MPI_Finalize();
	return 0;

} // end main()

//-- end of mpi_jet.c ---